"""Shared metaprogramming helpers for GECCO.
"""

import abc
import contextlib
import functools
import importlib
import locale
import operator
import typing
from multiprocessing.pool import Pool
from typing import Any, Callable, Iterable, List, Tuple, Optional, Type

if typing.TYPE_CHECKING:
    from types import TracebackType

    _S = typing.TypeVar("_S")
    _T = typing.TypeVar("_T")
    _A = typing.TypeVar("_A")
    _R = typing.TypeVar("_R")
    # _F = typing.TypeVar("_F", bound=Callable[[_A], _R])

class classproperty(property):
    """A class property decorator.

    Example:
        Use `classproperty` to create a property which counts the number of
        times it has been accessed:

        >>> class X():
        ...     __COUNT = 0
        ...     @classproperty
        ...     def count(cls):
        ...         cls.__COUNT += 1
        ...         return cls.__COUNT
        >>> X.count
        1
        >>> X.count
        2

    """

    def __init__(self, f: Callable[["_S"], "_T"]) -> None:
        self.f = f

    def __get__(self, obj: object, owner: "_S") -> "_T":  # type: ignore
        return self.f(owner)


class requires:
    """A decorator for functions that require optional dependencies.
    """

    def __init__(self, module_name):
        self.module_name = module_name

        try:
            self.module = importlib.import_module(module_name)
        except ImportError as err:
            self.module = err

    def __call__(self, func):

        if isinstance(self.module, ImportError):

            @functools.wraps(func)
            def newfunc(*args, **kwargs):
                msg = f"calling {func.__qualname__} requires module {self.module.name}"
                raise RuntimeError(msg) from self.module
        else:

            newfunc = func
            basename = self.module_name.split(".")[-1]
            newfunc.__globals__[basename] = self.module

        return newfunc


class OrderedPoolWrapper:
    """A `Pool` wrapper that returns results in the order they were given.
    """

    class _OrderedFunc:

        def __init__(self, inner: Callable[["_A"], "_R"], star: bool = False) -> None:
            self.inner = inner
            self.star = star

        def __call__(self, args: Tuple[int, "_A"]) -> Tuple[int, "_R"]:
            i, other = args
            if self.star:
                return i, self.inner(*other)  # type: ignore
            else:
                return i, self.inner(other)

    def __init__(self, inner: Pool) -> None:
        self.inner = inner

    def __enter__(self) -> "OrderedPoolWrapper":
        self.inner.__enter__()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional["TracebackType"],
    ) -> Optional[bool]:
        return self.inner.__exit__(exc_type, exc_value, traceback)

    def map(self, func: Callable[["_A"], "_R"], it: Iterable["_A"]) -> List["_R"]:
        wrapped_it = enumerate(it)
        wrapped_func = self._OrderedFunc(func)
        results = self.inner.map(wrapped_func, wrapped_it)
        results.sort(key=operator.itemgetter(0))
        return list(map(operator.itemgetter(1), results))

    def starmap(self, func: Callable[..., "_R"], it: Iterable[Iterable[Any]]) -> List["_R"]:
        wrapped_it = enumerate(it)
        wrapped_func = self._OrderedFunc(func, star=True)
        results = self.inner.map(wrapped_func, wrapped_it)
        results.sort(key=operator.itemgetter(0))
        return list(map(operator.itemgetter(1), results))


class UniversalContainer(object):
    """A container that contains everything.
    """

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __contains__(self, item: object) -> bool:
        return True


@contextlib.contextmanager
def patch_locale(name: str):
    """Create a context manager to locally change the locale in use.
    """
    lc = locale.setlocale(locale.LC_TIME)
    try:
        locale.setlocale(locale.LC_TIME, name)
        yield
    finally:
        locale.setlocale(locale.LC_TIME, lc)



import abc
import errno
import io
import subprocess
import typing
from typing import Iterable, Optional, Type, TextIO
from subprocess import DEVNULL

from ._meta import classproperty


class Dumpable(metaclass=abc.ABCMeta):
    """A metaclass for objects that can be dumped to a text file.
    """

    @abc.abstractmethod
    def dump(self, fh: TextIO) -> None:
        raise NotImplementedError

    def dumps(self) -> str:
        s = io.StringIO()
        self.dump(s)
        return s.getvalue()



"""Implementation of the main ``gecco`` command.
"""

import contextlib
import signal
import sys
import textwrap
import typing
import warnings
from typing import Mapping, Optional, Type

import docopt
import operator

from ... import __version__
from ..._meta import classproperty
from .._utils import in_context, patch_showwarnings
from . import __name__ as __parent__
from ._base import Command, CommandExit, InvalidArgument

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata


class Main(Command):
    """The *main* command launched before processing subcommands.
    """

    _entry_points_cache = None

    @classproperty
    def _entry_points(cls):
        if cls._entry_points_cache is None:
            cls._entry_points_cache = importlib_metadata.entry_points().get(__parent__, [])
        return cls._entry_points_cache

    @classmethod
    def _get_subcommand_names(cls) -> Mapping[str, Type[Command]]:
        return [cmd.name for cmd in cls._entry_points]

    @classmethod
    def _get_subcommands(cls) -> Mapping[str, Type[Command]]:
        commands = {}
        for cmd in cls._entry_points:
            try:
                commands[cmd.name] = cmd.load()
            except Exception:
                pass
        return commands

    @classmethod
    def _get_subcommand_by_name(cls, name: str) -> Optional[Type[Command]]:
        for cmd in cls._entry_points:
            if cmd.name == name:
                return cmd.load()
        return None

    # --

    @classmethod
    def doc(cls, fast=False):  # noqa: D102
        if fast:
            commands = (f"    {cmd}" for cmd in cls._get_subcommand_names())
        else:
            commands = (
                "    {:12}{}".format(name, typing.cast(Command, cmd).summary)
                for name, cmd in sorted(
                    cls._get_subcommands().items(), key=operator.itemgetter(0)
                )
            )
        return (
            textwrap.dedent(
                """
        gecco - Gene Cluster Prediction with Conditional Random Fields

        Usage:
            gecco [-v | -vv | -q | -qq] <cmd> [<args>...]
            gecco --version
            gecco --help [<cmd>]

        Commands:
        {commands}

        Parameters:
            -h, --help                 show the message for ``gecco`` or
                                       for a given subcommand.
            -q, --quiet                silence any output other than errors
                                       (-qq silences everything).
            -v, --verbose              increase verbosity (-v is verbose,
                                       -vv is very verbose and makes the
                                       output more suitable for logging).
            -V, --version              show the program version and exit.

        """
            )
            .lstrip()
            .format(commands="\n".join(commands))
        )

    _options_first = True

    # --

    def execute(self, ctx: contextlib.ExitStack) -> int:
        # Run the app, elegantly catching any interrupts or exceptions
        try:
            # check arguments and enter context
            self._check()
            ctx.enter_context(patch_showwarnings(self._showwarnings))

            # Get the subcommand class
            subcmd_name = self.args["<cmd>"]
            try:
                subcmd_cls = self._get_subcommand_by_name(subcmd_name)
            except ImportError as err:
                self._on_import_error(subcmd_name, err)
                return 1

            # exit if no known command was found
            if subcmd_name is not None and subcmd_cls is None:
                self.error("Unknown subcommand", repr(subcmd_name))
                return 1
            # if a help message was required, delegate to the `gecco help` command
            if (
                self.args["--help"]
                or "-h" in self.args["<args>"]
                or "--help" in self.args["<args>"]
            ):
                subcmd = typing.cast(Type[Command], self._get_subcommand_by_name("help"))(
                    argv=["help"] + [subcmd_name],
                    stream=self._stream,
                    options=self.args,
                    config=self.config,
                )
            # print version information if `--version` in flags
            elif self.args["--version"]:
                self.console.print("gecco", __version__)
                return 0
            # initialize the command if is valid
            else:
                subcmd = typing.cast(Type[Command], subcmd_cls)(
                    argv=[self.args["<cmd>"]] + self.args["<args>"],
                    stream=self._stream,
                    options=self.args,
                    config=self.config,
                )
                subcmd.verbose = self.verbose
                subcmd.quiet = self.quiet
                subcmd.progress.disable = self.quiet > 0
            # run the subcommand
            return subcmd.execute(ctx)
        except CommandExit as sysexit:
            return sysexit.code
        except KeyboardInterrupt:
            self.error("Interrupted")
            return -signal.SIGINT
        except ImportError as err:
            self._on_import_error(subcmd_name, err)
            return 1
        except Exception as e:
            import rich.traceback

            self.error(
                "An unexpected error occurred. Consider opening"
                " a new issue on the bug tracker"
                " ( https://github.com/zellerlab/GECCO/issues/new ) if"
                " it persists, including the traceback below:"
            )
            traceback = rich.traceback.Traceback.from_exception(type(e), e, e.__traceback__)
            self.console.print(traceback)
            # return errno if exception has any
            return typing.cast(int, getattr(e, "errno", 1))



"""Generic protocol for ORF detection in DNA sequences.
"""

import abc
import io
import itertools
import os
import queue
import threading
import tempfile
import typing
from multiprocessing.sharedctypes import Value
from typing import Callable, Iterable, Iterator, List, Optional

import Bio.SeqIO
import pyrodigal
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from .model import Gene, Protein, Strand

if typing.TYPE_CHECKING:
    from Bio.SeqRecord import SeqRecord


__all__ = ["ORFFinder", "PyrodigalFinder"]


class ORFFinder(metaclass=abc.ABCMeta):
    """An abstract base class to provide a generic ORF finder.
    """

    @abc.abstractmethod
    def find_genes(self, records: Iterable[SeqRecord]) -> Iterable[Gene]:  # type: ignore
        """Find all genes from a DNA sequence.
        """
        return NotImplemented  # type: ignore


class PyrodigalFinder(ORFFinder):
    """An `ORFFinder` that uses the Pyrodigal bindings to Prodigal.

    Prodigal is a fast and reliable protein-coding gene prediction
    for prokaryotic genomes, with support for draft genomes and metagenomes.

    See Also:
        `Doug Hyatt, Gwo-Liang Chen, Philip F. LoCascio, Miriam L. Land,
        Frank W. Larimer and Loren J. Hauser.
        "Prodigal: Prokaryotic Gene Recognition and Translation Initiation
        Site Identification", BMC Bioinformatics 11 (8 March 2010), p119
        <https://doi.org/10.1186/1471-2105-11-119>`_

    """

    class _Worker(threading.Thread):

        @staticmethod
        def _none_callback(record, found, total):
            pass

        def __init__(
            self,
            metagenome: bool,
            record_count: "Value",
            record_queue: "queue.Queue[typing.Optional[typing.Tuple[int, SeqRecord]]]",
            orfs_queue: "queue.Queue[Gene]",
            kill_switch: threading.Event,
            orfs_found: typing.List[threading.Event],
            callback: Optional[Callable[[SeqRecord, int], None]],
        ) -> None:
            super().__init__()
            self.pyrodigal = pyrodigal.Pyrodigal(meta=metagenome)
            self.record_queue = record_queue
            self.record_count = record_count
            self.orfs_queue = orfs_queue
            self.orfs_found = orfs_found
            self.callback = callback or self._none_callback
            self.error: typing.Optional[BaseException] = None
            self.kill_switch = kill_switch

        def kill(self) -> None:
            self.kill_switch.set()

        def run(self) -> None:
            while not self.kill_switch.is_set():
                args = self.record_queue.get()
                if args is None:
                    self.record_queue.task_done()
                    return
                else:
                    index, record = args
                try:
                    self.process(index, record)
                    self.record_queue.task_done()
                except BaseException as exc:
                    self.error = exc
                    self.kill()
                    return
                finally:
                    self.orfs_found[index].set()

        def process(self, index, record):
            orfs = self.pyrodigal.find_genes(str(record.seq))
            genes = []
            for j, orf in enumerate(orfs):
                # wrap the protein into a Protein object
                prot_seq = Seq(orf.translate())
                protein = Protein(id=f"{record.id}_{j+1}", seq=prot_seq)
                # wrap the gene into a Gene
                genes.append(Gene(
                    source=record,
                    start=min(orf.begin, orf.end),
                    end=max(orf.begin, orf.end),
                    strand=Strand.Coding if orf.strand == 1 else Strand.Reverse,
                    protein=protein,
                    qualifiers={
                        "inference": ["ab initio prediction:Prodigal:2.6"],
                        "transl_table": str(orf.translation_table),
                    }
                ))
            self.orfs_queue.put((index, genes))
            self.callback(record, len(orfs), self.record_count.value)


    def __init__(self, metagenome: bool = True, cpus: int = 0) -> None:
        """Create a new `PyrodigalFinder` instance.

        Arguments:
            metagenome (bool): Whether or not to run PRODIGAL in metagenome
                mode, defaults to `True`.
            cpus (int): The number of threads to use to run Pyrodigal in
                parallel. Pass ``0`` to use the number of CPUs on the machine.

        """
        super().__init__()
        self.metagenome = metagenome
        self.cpus = cpus

    def find_genes(
        self,
        records: Iterable[SeqRecord],
        progress: Optional[Callable[[SeqRecord, int], None]] = None,
    ) -> Iterator[Gene]:
        """Find all genes contained in a sequence of DNA records.

        Arguments:
            records (iterable of `~Bio.SeqRecord.SeqRecord`): An iterable of
                DNA records in which to find genes.
            progress (callable, optional): A progress callback of signature
                ``progress(record, total)`` that will be called everytime a
                record has been processed successfully, with ``record`` being
                the `~Bio.SeqRecord.SeqRecord` instance, and ``total`` being
                the total number of records to process.

        Yields:
            `~gecco.model.Gene`: An iterator over all the genes found in
            the given records.

        """
        # count the number of CPUs to use
        _cpus = self.cpus if self.cpus > 0 else os.cpu_count()

        # create the queue to pass the objects around
        orfs_found: "typing.List[threading.Event]" = []
        record_count = Value('i')
        record_queue = typing.cast("queue.Queue[typing.Optional[typing.Tuple[int, SeqRecord]]]", queue.Queue())
        orfs_queue = typing.cast("queue.PriorityQueue[typing.Tuple[int, SeqRecord]]", queue.PriorityQueue())
        kill_switch = threading.Event()

        # create and launch one thread per CPU
        threads = []
        for _ in range(_cpus):
            thread = self._Worker(
                self.metagenome,
                record_count,
                record_queue,
                orfs_queue,
                kill_switch,
                orfs_found,
                callback=progress,
            )
            thread.start()
            threads.append(thread)

        # catch exceptions to kill threads in the background before exiting
        try:
            # enumerate queries, so that we know the index of each query
            # and we can yield the results in the same order
            queries = enumerate(records)
            # initially feed one query per thread so that they can start
            # working before we enter the main loop
            for index, record in itertools.islice(queries, _cpus):
                record_count.value += 1
                orfs_found.append(threading.Event())
                record_queue.put((index, record))
            # alternate between feeding queries to the threads and
            # yielding back results, if available
            orfs_yielded = 0
            while orfs_yielded < record_count.value:
                # get the next query record
                try:
                    index, query = next(queries)
                    record_count.value += 1
                    orfs_found.append(threading.Event())
                    record_queue.put((index, query))
                except StopIteration:
                    break
                # yield the top hits for the next query, if available
                if orfs_found[orfs_yielded].is_set():
                    yield from orfs_queue.get_nowait()[1]
                    orfs_yielded += 1
            # now that we exhausted all queries, poison pill the
            # threads so they stop on their own
            for _ in threads:
                record_queue.put(None)
            # yield remaining results
            while orfs_yielded < record_count.value:
                orfs_found[orfs_yielded].wait()
                yield from orfs_queue.get_nowait()[1]
                orfs_yielded += 1
        except queue.Empty:
            # the only way we can get queue.Empty is if a thread has set
            # the flag for `orfs_found[i]` without actually putting it in
            # the queue: this only happens when a background thread raised
            # an exception, so we must chain it
            for thread in threads:
                if thread.error is not None:
                    raise thread.error from None
        except BaseException:
            # make sure threads are killed to avoid being stuck,
            # e.g. after a KeyboardInterrupt
            kill_switch.set()
            raise



"""BGC prediction using a conditional random field.
"""

import csv
import copy
import functools
import hashlib
import itertools
import io
import math
import numbers
import operator
import os
import pickle
import random
import textwrap
import typing
import warnings
from multiprocessing.pool import Pool
from typing import (
    Callable,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

import numpy
import tqdm
import sklearn_crfsuite
import sklearn.model_selection
import sklearn.preprocessing

from ..model import Gene
from .._meta import OrderedPoolWrapper
from . import features
from .cv import LeaveOneGroupOut
from .select import fisher_significance

try:
    import importlib.resources as importlib_resources
except ImportError:
    import importlib_resources

__all__ = ["ClusterCRF"]


class ClusterCRF(object):
    """A wrapper for `sklearn_crfsuite.CRF` to work with the GECCO data model.
    """

    @classmethod
    def trained(cls, model_path: Optional[str] = None) -> "ClusterCRF":
        """Create a new pre-trained `ClusterCRF` instance from a model path.

        Arguments:
            model_path (`str`, optional): The path to the model directory
                obtained with the ``gecco train`` command. If `None` given,
                use the embedded model.

        Returns:
            `~gecco.crf.ClusterCRF`: A CRF model that can be used to perform
            predictions without training first.

        Raises:
            `ValueError`: If the model data does not match its hash.

        """
        # get the path to the pickled model and read its signature file
        if model_path is not None:
            pkl_file = open(os.path.join(model_path, "model.pkl"), "rb")
            md5_file = open(os.path.join(model_path, "model.pkl.md5"))
        else:
            pkl_file = importlib_resources.open_binary(__name__, "model.pkl")
            md5_file = importlib_resources.open_text(__name__, "model.pkl.md5")
        with md5_file as sig:
            signature = sig.read().strip()

        # check the file content matches its MD5 hashsum
        hasher = hashlib.md5()
        with pkl_file as bin:
            read = functools.partial(bin.read, io.DEFAULT_BUFFER_SIZE)
            for chunk in iter(read, b""):
                hasher.update(typing.cast(bytes, chunk))
            if hasher.hexdigest().upper() != signature.upper():
                raise ValueError("MD5 hash of model data does not match signature")
            pkl_file.seek(0)
            return pickle.load(bin)  # type: ignore

    def __init__(
        self,
        feature_type: str = "protein",
        algorithm: str = "lbfgs",
        overlap: int = 2,
        pool_factory: Union[Type[Pool], Callable[[Optional[int]], Pool]] = Pool,
        **kwargs: Dict[str, object],
    ) -> None:
        """Create a new `ClusterCRF` instance.

        Arguments:
            feature_type (`str`): Defines how features should be extracted.
                Should be either *domain*, *protein*, or *overlap*.
            algorithm (`str`): The optimization algorithm for the model. See
                https://sklearn-crfsuite.readthedocs.io/en/latest/api.html
                for available values.
            overlap (`int`): In case of ``feature_type = "overlap"``, defines
                the sliding window size to use. The resulting window width is
                ``2*overlap+1``.
            pool_factory (`multiprocessing.pool.Pool` subclass, or callable):
                The factory to use to create a new pool instance for methods
                that can perform operations in parallel. *It is called with
                a single argument which is the number of workers to create,
                or* `None` *to create a much workers as there are CPUs.*

        Any additional keyword argument is passed as-is to the internal
        `~sklearn_crfsuite.CRF` constructor.

        Raises:
            `ValueError`: if ``feature_type`` has an invalid value.
            `TypeError`: if one of the ``*_columns`` argument is not iterable.

        """
        if feature_type not in {"single", "overlap", "group"}:
            raise ValueError(f"unexpected feature type: {feature_type!r}")

        self.feature_type: str = feature_type
        self.overlap: int = overlap
        self.algorithm = algorithm
        self.pool_factory = pool_factory
        self.significance: Optional[Dict[str, float]] = None
        self.significant_features: Optional[FrozenSet[str]] = None
        self.model = sklearn_crfsuite.CRF(
            algorithm=algorithm,
            all_possible_transitions=True,
            all_possible_states=True,
            **kwargs,
        )

    def predict_probabilities(self, genes: Iterable[Gene], *, cpus: Optional[int] = None) -> List[Gene]:
        """Predict how likely each given gene is part of a gene cluster.
        """
        _cpus = os.cpu_count() if not cpus else cpus
        # group input genes by sequence
        groups = itertools.groupby(genes, key=operator.attrgetter("source.id"))
        seqs = [sorted(group, key=operator.attrgetter("start")) for _, group in groups]

        # select the feature extraction method
        if self.feature_type == "group":
            extract = features.extract_features_group
            annotate = features.annotate_probabilities_group
        elif self.feature_type == "single":
            extract = features.extract_features_single
            annotate = features.annotate_probabilities_single
        elif self.feature_type == "overlap":
            raise NotImplementedError("todo: `features.extract_features_overlap`")
        else:
            raise ValueError("invalid feature type")

        # proces each sequence / group in parallel
        with OrderedPoolWrapper(self.pool_factory(_cpus)) as pool:
            # extract features in parallel and predict cluster probabilities
            marginals = self.model.predict_marginals(pool.map(extract, seqs))
            # Annotate the genes with the predicted probabilities
            annotated_seqs = pool.starmap(annotate, zip(seqs, marginals))

        # return the genes that were passed as input but now having BGC
        # probabilities set
        return list(itertools.chain.from_iterable(annotated_seqs)) # type: ignore

    def fit(
        self,
        genes: Iterable[Gene],
        select: Optional[float] = None,
        shuffle: bool = True,
        *,
        cpus: Optional[int] = None,
        correction_method: Optional[str] = None,
    ) -> None:
        _cpus = os.cpu_count() if not cpus else cpus
        # select the feature extraction method
        if self.feature_type == "group":
            extract_features = features.extract_features_group
            extract_labels = features.extract_labels_group
        elif self.feature_type == "single":
            extract_features = features.extract_features_single
            extract_labels = features.extract_labels_single
        elif self.feature_type == "overlap":
            raise NotImplementedError("todo: `features.extract_features_overlap`")
        else:
            raise ValueError("invalid feature type")

        # group input genes by sequence
        groups = itertools.groupby(genes, key=operator.attrgetter("source.id"))
        seqs = [sorted(group, key=operator.attrgetter("start")) for _, group in groups]

        # shuffle sequences
        if shuffle:
            random.shuffle(seqs)

        # perform feature selection
        if select is not None:
            if select <= 0 or select > 1:
                raise ValueError(f"invalid value for select: {select}")
            # find most significant features
            self.significance = sig = fisher_significance(
                (g.protein for seq in seqs for g in seq),
                correction_method=correction_method,
            )
            sorted_sig = sorted(sig, key=sig.get)[:int(select*len(sig))]
            self.significant_features = frozenset(sorted_sig)
            # check that we don't select "random" domains
            if sig[sorted_sig[-1]] == 1.0:
                warnings.warn(
                    "Selected features still include domains with a p-value "
                    "of 1, consider reducing the selected fraction.",
                    UserWarning
                )
            # remove non significant domains
            for i, seq in enumerate(seqs):
                for j, gene in enumerate(seq):
                    seqs[i][j] = gene = copy.deepcopy(gene)
                    gene.protein.domains = [
                        domain for domain in gene.protein.domains
                        if domain.name in self.significant_features
                    ]

        # proces each sequence / group in parallel
        with OrderedPoolWrapper(self.pool_factory(_cpus)) as pool:
            # extract features in parallel and predict cluster probabilities
            X = pool.map(extract_features, seqs)
            Y = pool.map(extract_labels, seqs)

        # check labels
        if all(y == "1" for y in Y):
            raise ValueError("only positives labels found, something is wrong.")
        elif all(y == "0" for y in Y):
            raise ValueError("only negative labels found, something is wrong.")

        # fit the model
        self.model.fit(X, Y)



"""Supervised classifier to predict the biosynthetic type of a cluster.
"""

import csv
import functools
import operator
import os
import typing
from typing import Callable, Dict, List, Iterable, Optional, Sequence, Tuple, Union

import numpy
import scipy.sparse
import sklearn.ensemble
import sklearn.preprocessing

from ..model import ProductType, Cluster

try:
    import importlib.resources as importlib_resources
except ImportError:
    import importlib_resources

__all__ = ["TypeBinarizer", "TypeClassifier"]


class TypeBinarizer(sklearn.preprocessing.MultiLabelBinarizer):
    """A `MultiLabelBinarizer` working with `ProductType` instances.
    """

    def __init__(self):
        self.classes_ = sorted(x for x in ProductType.__members__.values() if x)
        super().__init__(classes=self.classes_)

    def transform(self, y: Iterable[ProductType]) -> Iterable[Iterable[int]]:
        matrix = numpy.zeros((len(y), len(self.classes_)))
        for i, label in enumerate(y):
            for j, cls in enumerate(self.classes_):
                matrix[i, j] = bool(label & cls)
        return matrix

    def inverse_transform(self, yt: Iterable[Iterable[int]]) -> Iterable[ProductType]:
        classes = []
        for y in yt:
            filtered = [cls for i, cls in enumerate(self.classes_) if y[i]]
            classes.append(ProductType.pack(filtered))
        return classes


class TypeClassifier(object):
    """A wrapper to predict the type of a `~gecco.model.Cluster`.
    """

    @classmethod
    def trained(cls, model_path: Optional[str] = None) -> "TypeClassifier":
        """Create a new `TypeClassifier` pre-trained with embedded data.

        Arguments:
            model_path (`str`, optional): The path to the model directory
                obtained with the ``gecco train`` command. If `None` given,
                use the embedded training data.

        Returns:
            `~gecco.types.TypeClassifier`: A random forest model that can be
            used to perform BGC type predictions without training first.

        """

        if model_path is not None:
            doms_file = open(os.path.join(model_path, "domains.tsv"))
            typs_file = open(os.path.join(model_path, "types.tsv"))
            comp_file = open(os.path.join(model_path, "compositions.npz"), "rb")
        else:
            doms_file = importlib_resources.open_text(__name__, "domains.tsv")
            typs_file = importlib_resources.open_text(__name__, "types.tsv")
            comp_file = importlib_resources.open_binary(__name__, "compositions.npz")

        with comp_file as comp_src:
            compositions = scipy.sparse.load_npz(comp_src)
        with doms_file as doms_src:
            domains = [ line.strip() for line in doms_src ]
        with typs_file as typs_src:
            types = [
                ProductType.pack(ProductType.__members__[ty] for ty in raw.split(";"))
                for raw in (line.split("\t")[1].strip() for line in typs_src)
            ]

        classifier = cls(random_state=0)
        types_bin = classifier.binarizer.transform(types)
        classifier.model.fit(compositions, y=types_bin)
        classifier.model.attributes_ = domains
        return classifier

    def __init__(self, **kwargs: object) -> None:
        """Instantiate a new type classifier.

        Keyword Arguments:
            Any additional keyword argument is passed as argument to the
            internal `~sklearn.ensemble.RandomForestClassifier` constructor.

        """
        self.model = sklearn.ensemble.RandomForestClassifier(**kwargs)
        self.binarizer = TypeBinarizer()

    _S = typing.TypeVar("_S", bound=Sequence["Cluster"])

    def predict_types(self, clusters: "_S") -> "_S":
        """Predict types for each of the given clusters.
        """
        # extract domain compositions from input clusters
        comps = numpy.array([c.domain_composition(self.model.attributes_) for c in clusters])

        # predict type probabilities with the internal classifier
        probas = self.model.predict_proba(comps)

        # extract only the *positive* probabilites and translate them to proper
        # type predictions using the binarizer
        if len(comps) == 1:
            posit = numpy.array([[1 - cls[0][0] for cls in probas]])
        else:
            posit = 1 - numpy.array(probas)[:, :, 0].transpose()

        # translate probabilities into product type predictions
        types = self.binarizer.inverse_transform(posit > 0.5)

        # annotate the input clusters
        results = zip(typing.cast(Iterable["Cluster"], clusters), posit, types)
        for cluster, proba, ty in results:
            cluster.type = ty
            cluster.type_probabilities = dict(zip(self.binarizer.classes_, proba))
        return clusters



"""Features extraction and sanitization utilities for `~gecco.crf.ClusterCRF`.
"""
import itertools
import typing
from typing import Iterable, List, Sequence, Dict

from ..model import Gene

if typing.TYPE_CHECKING:
    _S = typing.TypeVar("_S", bound=Sequence[Gene])


def extract_features_group(sequence: Iterable[Gene]) -> List[Dict[str, float]]:
    """Extract features at the group/protein level from an iterable of genes.
    """
    # FIXME: currently (v0.3.0 and later) we have to hide proteins missing
    # domain annotations because the model has only been trained with proteins
    # that had domains (since unannotated proteins do not appear in the input
    # feature table used for training)
    #
    #        Fixing this requires retraining the model so that it is aware of
    #  unannotated proteins in the training data, and learns to ignore them.
    #  When it is done, make sure to edit `postprocessing` as well.
    return [{d.name: 1 - d.i_evalue for d in g.protein.domains } for g in sequence if g.protein.domains]

def extract_labels_group(sequence: Iterable[Gene]) -> List[str]:
    """Extract labels at the group/protein level from an iterable of genes.
    """
    return [
        str(int(g.average_probability > 0.5))  # type: ignore
        for g in sequence
        if g.protein.domains
    ]


def annotate_probabilities_group(sequence: "_S", marginals: List[Dict[str, float]]) -> "_S":
    """Annotate genes with marginals obtained from a CRF at the protein level.
    """
    probabilities = iter(marginals)
    for gene in sequence:
        if not gene.protein.domains:
            continue
        p = next(probabilities)["1"]
        gene.protein.domains = [ d.with_probability(p) for d in gene.protein.domains]
    return sequence


def annotate_probabilities_single(sequence: "_S", marginals: List[Dict[str, float]]) -> "_S":
    """Annotate genes with marginals obtained from a CRF at the domain level.
    """
    domains = [ domain for gene in sequence for domain in gene.protein.domains ]
    for domain, probability in itertools.zip_longest(domains, marginals):
        assert domain is not None
        assert probability is not None
        domain.probability = probability["1"]
    return sequence

def extract_labels_single(sequence: Iterable[Gene]) -> List[str]:
    """Extract labels at the domain level from an iterable of genes.
    """
    return [
        str(int(d.probability > 0.5))  # type: ignore
        for g in sequence
        for d in g.protein.domains
    ]


def extract_features_single(sequence: Iterable[Gene]) -> List[Dict[str, float]]:
    """Extract features at the domain level from an iterable of genes.
    """
    return [{d.name: 1 - d.i_evalue} for g in sequence for d in g.protein.domains]



"""Data layer classes storing information needed for BGC detection.
"""

import collections
import csv
import datetime
import enum
import functools
import itertools
import math
import operator
import re
import typing
from array import array
from collections import OrderedDict
from collections.abc import Sized
from dataclasses import dataclass, field
from typing import Dict, Iterable, List, Mapping, Optional, Sequence, TextIO, NamedTuple, Union, Iterator

import Bio
import numpy
from Bio.Seq import Seq
from Bio.SeqFeature import SeqFeature, FeatureLocation, CompoundLocation, Reference
from Bio.SeqRecord import SeqRecord

from . import __version__
from ._base import Dumpable
from ._meta import requires, patch_locale


__all__ = [
    "ProductType",
    "Strand",
    "Domain",
    "Protein",
    "Gene",
    "Cluster",
    "FeatureTable",
    "ClusterTable"
]


# fmt: off
class ProductType(enum.IntFlag):
    """A flag to declare the type of product synthesized by a gene cluster.
    """

    Unknown    = 0b00000000
    Other      = 0b00000001
    Alkaloid   = 0b00000010
    Polyketide = 0b00000100
    RiPP       = 0b00001000
    Saccharide = 0b00010000
    Terpene    = 0b00100000
    NRP        = 0b01000000

    @classmethod
    def pack(cls, members: Iterable["ProductType"]) -> "ProductType":
        """Pack together a list of individual product types.

        Example:
            >>> types = [ProductType.Polyketide, ProductType.Saccharide]
            >>> ProductType.pack(types)
            <ProductType.Saccharide|Polyketide: 20>

        """
        return functools.reduce(operator.or_, members, cls.Unknown)

    def unpack(self) -> List["ProductType"]:
        """Unpack a composite `ProductType` into a list of individual types.

        Example:
            >>> ty = ProductType.Polyketide | ProductType.Saccharide
            >>> ty.unpack()
            [<ProductType.Polyketide: 4>, <ProductType.Saccharide: 16>]

        """
        return [ x for x in ProductType.__members__.values() if (x & self) ]


class Strand(enum.IntEnum):
    """A flag to declare on which DNA strand a gene is located.
    """

    Coding = 1
    Reverse = -1

    @property
    def sign(self) -> str:
        """`str`: The strand as a single sign (``+`` or ``-``).
        """
        return "+" if self is Strand.Coding else "-"


@dataclass(frozen=True)
class Domain:
    """A conserved region within a protein.

    Attributes:
        name (`str`): The accession of the protein domain in the source HMM.
        start (`int`): The start coordinate of the domain within the protein
            sequence (first amino-acid at 1).
        end (`int`): The end coordinate of the domain within the protein
            sequence (inclusive).
        hmm (`str`): The name of the HMM library this domain belongs to
            (e.g. ``Pfam``, ``Panther``).
        i_evalue (`float`): The independent e-value reported by ``hmmsearch``
            that measures how reliable the domain annotation is.
        pvalue (`float`): The p-value reported by ``hmmsearch`` that measure
            how likely the domain score is.
        probability (`float`, optional): The probability that this domain
            is part of a BGC, or `None` if no prediction has been made yet.
        qualifiers (`dict`, optional): A dictionary of feature qualifiers that
            is added to the `~Bio.SeqFeature.SeqFeature` built from this
            `Domain`.

    """

    name: str
    start: int
    end: int
    hmm: str
    i_evalue: float
    pvalue: float
    probability: Optional[float] = None
    qualifiers: Mapping[str, Union[str, List[str]]] = field(default_factory=dict)

    def with_probability(self, probability: Optional[float]) -> "Domain":
        """Copy the current domain and assign it a BGC probability.
        """
        return Domain(
            self.name, self.start, self.end, self.hmm, self.i_evalue, self.pvalue,
            probability, self.qualifiers
        )

    def to_seq_feature(self, protein_coordinates: bool = False) -> SeqFeature:
        """Convert the domain to a single feature.

        Arguments:
            protein_coordinates (`bool`): Set to `True` for the feature
                coordinates to be given in amino-acids, or to `False` in
                nucleotides.

        """
        stride = 1 if protein_coordinates else 3
        loc = FeatureLocation((self.start-1)*stride, self.end*stride)
        qualifiers = dict(self.qualifiers)
        qualifiers.setdefault("standard_name", self.name)
        return SeqFeature(location=loc, type="misc_feature", qualifiers=qualifiers)


@dataclass
class Protein:
    """A sequence of amino-acids translated from a gene.

    Attributes:
        id (`str`): The identifier of the protein.
        seq (`~Bio.Seq.Seq`): The sequence of amino-acids of this protein.
        domains (`list` of `~gecco.model.Domain`): A list of domains found
            in the protein sequence.

    """

    id: str
    seq: Seq
    domains: List[Domain] = field(default_factory=list)

    def to_seq_record(self) -> SeqRecord:
        """Convert the protein to a single record.
        """
        # FIXME: add domains
        record = SeqRecord(self.seq, id=self.id, name=self.id)

        biopython_version = tuple(map(int, Bio.__version__.split(".")))
        if biopython_version < (1, 77):
            from Bio import Alphabet

            record.seq.alphabet = Alphabet.generic_protein

        return record


@dataclass
class Gene:
    """A nucleotide sequence coding a protein.

    Attributes:
        source (`~Bio.SeqRecord.SeqRecord`): The DNA sequence this gene was
            found in, as a Biopython record.
        start (`int`): The index of the leftmost nucleotide of the gene within
            the source sequence, independent of the strandedness.
        end (`int`): The index of the rightmost nucleotide of the gene within
            the source sequence.
        strand (`~gecco.model.Strand`): The strand where the gene is located.
        protein (`~gecco.model.Protein`): The protein translated from this
            gene.
        qualifiers (`dict`, optional): A dictionary of feature qualifiers that
            is added to the `~Bio.SeqFeature.SeqFeature` built from this
            `Gene`.

    """

    source: SeqRecord
    start: int
    end: int
    strand: Strand
    protein: Protein
    qualifiers: Mapping[str, Union[str, List[str]]] = field(default_factory=dict)

    @property
    def id(self) -> str:
        """`str`: The identifier of the gene (same as the protein identifier).
        """
        return self.protein.id

    @property
    def average_probability(self) -> Optional[float]:
        """`float`: The average of domain probabilities of being biosynthetic.
        """
        p = [d.probability for d in self.protein.domains if d.probability is not None]
        return sum(p) / len(p) if p else None

    @property
    def maximum_probability(self) -> Optional[float]:
        """`float`: The highest of domain probabilities of being biosynthetic.
        """
        p = [d.probability for d in self.protein.domains if d.probability is not None]
        return max(p) if p else None

    # ---

    def to_seq_feature(self) -> SeqFeature:
        """Convert the gene to a single feature.
        """
        # NB(@althonos): we use inclusive 1-based ranges in the data model
        # but Biopython expects 0-based ranges with exclusive ends
        loc = FeatureLocation(start=self.start, end=self.end+1, strand=int(self.strand))
        qualifiers = dict(self.qualifiers)
        qualifiers.setdefault("locus_tag", self.protein.id)
        qualifiers.setdefault("translation", str(self.protein.seq))
        return SeqFeature(location=loc, type="CDS", qualifiers=qualifiers)


@dataclass
class Cluster:
    """A sequence of contiguous genes with biosynthetic activity.

    Attributes:
        id (`str`): The identifier of the gene cluster.
        genes (`list` of `~gecco.model.Gene`): A list of the genes belonging
            to this gene cluster.
        types (`gecco.model.ProductType`): The putative types of product
            synthesized by this gene cluster, according to similarity in
            domain composition with curated clusters.
        types_probabilities (`list` of `float`): The probability with which
            each BGC type was identified (same dimension as the ``types``
            attribute).

    """

    id: str
    genes: List[Gene]
    type: ProductType
    type_probabilities: Mapping[ProductType, float]

    def __init__(
        self,
        id: str,
        genes: Optional[List[Gene]] = None,
        type: ProductType = ProductType.Unknown,
        type_probabilities: Optional[Dict[ProductType, float]] = None,
    ):  # noqa: D107
        self.id = id
        self.genes = genes or list()
        self.type = type
        self.type_probabilities = type_probabilities or dict()

        # if len(self.types) != len(self.types_probabilities):
        #     err = "type and type probability lists must have the same dimensions"
        #     raise ValueError(err)


    @property
    def source(self) -> SeqRecord:  # type: ignore
        """`~Bio.SeqRecord.SeqRecord`: The sequence this cluster was found in.
        """
        return self.genes[0].source

    @property
    def start(self) -> int:
        """`int`: The start of this cluster in the source sequence.
        """
        return min(gene.start for gene in self.genes)

    @property
    def end(self) -> int:
        """`int`: The end of this cluster in the source sequence.
        """
        return max(gene.end for gene in self.genes)

    @property
    def average_probability(self) -> Optional[float]:
        """`float`: The average of proteins probability of being biosynthetic.
        """
        p = [g.average_probability for g in self.genes if g.average_probability is not None]
        return sum(p) / len(p) if p else None

    @property
    def maximum_probability(self) -> Optional[float]:
        """`float`: The highest of proteins probability of being biosynthetic.
        """
        p = [g.maximum_probability for g in self.genes if g.maximum_probability is not None]
        return max(p) if p else None

    # ---

    def domain_composition(
        self,
        all_possible: Optional[Sequence[str]] = None,
        normalize: bool = True,
        minlog_weights: bool = False,
        pvalue: bool = True,
    ) -> numpy.ndarray:
        """Compute weighted domain composition with respect to ``all_possible``.

        Arguments:
            all_possible (sequence of `str`, optional): A sequence containing
                all domain names to consider when computing domain composition
                for the BGC. If `None` given, then only domains within the
                cluster are taken into account.
            normalize (`bool`): Normalize the composition vector so that it
                sums to 1.
            minlog_weights (`bool`): Compute weight for each domain as
                :math:`-log_10(v)` (where :math:`v` is either the ``pvalue``
                or the ``i_evalue``, depending on the value of ``normalize``).
                Use :math:`1 - v` otherwise.
            pvalue (`bool`): Compute composition weights using the ``pvalue``
                of each domain, instead of the ``i_evalue``.

        Returns:
            `~numpy.ndarray`: A numerical array containing the relative domain
            composition of the BGC.

        """
        domains = [d for gene in self.genes for d in gene.protein.domains]
        names = numpy.array([domain.name for domain in domains])
        field = operator.attrgetter("pvalue" if pvalue else "i_evalue")
        if minlog_weights:
            weights = numpy.array([- math.log10(field(domain)) for domain in domains])
        else:
            weights = numpy.array([1 - field(domain) for domain in domains])

        unique_names = set(names)
        if all_possible is None:
            all_possible = numpy.unique(names)
        composition = numpy.zeros(len(all_possible))
        for i, dom in enumerate(all_possible):
            if dom in unique_names:
                composition[i] = numpy.sum(weights[names == dom])
        if normalize:
            return composition / (composition.sum() or 1)  # type: ignore
        return composition

    # ---

    def to_seq_record(self) -> SeqRecord:
        """Convert the cluster to a single record.

        Annotations of the source sequence are kept intact if they don't
        overlap with the cluster boundaries. Component genes are added on the
        record as *CDS* features. Annotated protein domains are added as
        *misc_feature*.

        """
        # store time of record creation
        now = datetime.datetime.now()

        # NB(@althonos): we use inclusive 1-based ranges in the data model
        # but slicing expects 0-based ranges with exclusive ends
        bgc = self.source[self.start - 1 : self.end]
        bgc.id = bgc.name = self.id

        # copy sequence annotations
        bgc.annotations = self.source.annotations.copy()
        bgc.annotations["topology"] = "linear"
        bgc.annotations["molecule_type"] = "DNA"
        with patch_locale("C"):
            bgc.annotations['date'] = now.strftime("%d-%b-%Y").upper()

        biopython_version = tuple(map(int, Bio.__version__.split(".")))
        if biopython_version < (1, 77):
            from Bio import Alphabet

            bgc.seq.alphabet = Alphabet.generic_dna

        # add GECCO preprint as a reference
        ref = Reference()
        ref.title = "Accurate de novo identification of biosynthetic gene clusters with GECCO"
        ref.journal = "bioRxiv (2021.05.03.442509)"
        ref.comment = "doi:10.1101/2021.05.03.442509"
        ref.authors = ", ".join([
            "Laura M Carroll",
            "Martin Larralde",
            "Jonas Simon Fleck",
            "Ruby Ponnudurai",
            "Alessio Milanese",
            "Elisa Cappio Barazzone",
            "Georg Zeller"
        ])
        bgc.annotations.setdefault("references", []).append(ref)

        # add GECCO-specific annotations as a structured comment
        structured_comment = bgc.annotations.setdefault("structured_comment", OrderedDict())
        structured_comment['GECCO-Data'] = {
            "version": f"GECCO v{__version__}",
            "creation_date": now.isoformat(),
            "biosyn_class": ";".join(sorted(ty.name for ty in self.type.unpack())),
            "alkaloid_probability": self.type_probabilities.get(ProductType.Alkaloid, 0.0),
            "polyketide_probability": self.type_probabilities.get(ProductType.Polyketide, 0.0),
            "ripp_probability": self.type_probabilities.get(ProductType.RiPP, 0.0),
            "saccharide_probability": self.type_probabilities.get(ProductType.Saccharide, 0.0),
            "terpene_probability": self.type_probabilities.get(ProductType.Terpene, 0.0),
            "nrp_probability": self.type_probabilities.get(ProductType.NRP, 0.0),
            "other_probability": self.type_probabilities.get(ProductType.Other, 0.0),
        }

        # add proteins as CDS features
        for gene in self.genes:
            # write gene as a /cds GenBank record
            cds = gene.to_seq_feature()
            cds.location += -self.start
            bgc.features.append(cds)
            # write domains as /misc_feature annotations
            for domain in gene.protein.domains:
                misc = domain.to_seq_feature(protein_coordinates=False)
                misc.location += cds.location.start
                bgc.features.append(misc)

        # return the complete BGC
        return bgc


class _UnknownSeq(Seq):
    """An unknown sequence that uses limited memory.

    Used by `FeatureTable.to_genes` to fake the `Gene.source.seq` attribute.
    """

    def __init__(self) -> None:
        super().__init__(data="")

    @typing.overload
    def __getitem__(self, index: int) -> str:
        pass

    @typing.overload
    def __getitem__(self, index: slice) -> Seq:
        pass

    def __getitem__(self, index: Union[slice, int]) -> Union[str, Seq]:
        if isinstance(index, slice):
            return Seq("N" * ((index.stop - index.start) // (index.step or 1)) )
        return "N"


@dataclass(frozen=True)
class FeatureTable(Dumpable, Sized):
    """A table storing condensed domain annotations from different genes.
    """

    sequence_id: List[str] = field(default_factory = list)
    protein_id: List[str] = field(default_factory = list)
    start: List[int] = field(default_factory = lambda: array("l"))          # type: ignore
    end: List[int] = field(default_factory = lambda: array("l"))            # type: ignore
    strand: List[str] = field(default_factory = list)
    domain: List[str] = field(default_factory = list)
    hmm: List[str] = field(default_factory = list)
    i_evalue: List[float] = field(default_factory = lambda: array("d"))     # type: ignore
    pvalue: List[float] = field(default_factory = lambda: array("d"))       # type: ignore
    domain_start: List[int] = field(default_factory = lambda: array("l"))   # type: ignore
    domain_end: List[int] = field(default_factory = lambda: array("l"))     # type: ignore
    bgc_probability: List[Optional[float]] = field(default_factory = list)

    class Row(NamedTuple):
        """A single row in a feature table.
        """

        sequence_id: str
        protein_id: str
        start: int
        end: int
        strand: str
        domain: str
        hmm: str
        i_evalue: float
        pvalue: float
        domain_start: int
        domain_end: int
        bgc_probability: Optional[float]

    @classmethod
    def from_genes(cls, genes: Iterable[Gene]) -> "FeatureTable":
        """Create a new feature table from an iterable of genes.
        """
        table = cls()
        for gene in genes:
            for domain in gene.protein.domains:
                table.sequence_id.append(gene.source.id)
                table.protein_id.append(gene.protein.id)
                table.start.append(gene.start)
                table.end.append(gene.end)
                table.strand.append(gene.strand.sign)
                table.domain.append(domain.name)
                table.hmm.append(domain.hmm)
                table.i_evalue.append(domain.i_evalue)
                table.pvalue.append(domain.pvalue)
                table.domain_start.append(domain.start)
                table.domain_end.append(domain.end)
                table.bgc_probability.append(domain.probability)
        return table

    def to_genes(self) -> Iterable[Gene]:
        """Convert a feature table to actual genes.

        Since the source sequence cannot be known, a *dummy* sequence is
        built for each gene of size ``gene.end``, so that each gene can still
        be converted to a `~Bio.SeqRecord.SeqRecord` if needed.
        """
        # group rows by protein/gene ID
        protein_indices = collections.defaultdict(list)
        for i, protein_id in enumerate(self.protein_id):
            protein_indices[protein_id].append(i)
        # yield genes in order
        for protein_id in sorted(protein_indices):
            rows = [self[i] for i in protein_indices[protein_id]]
            assert all(x.sequence_id == rows[0].sequence_id for x in rows)
            assert all(x.protein_id == rows[0].protein_id for x in rows)
            assert all(x.start == rows[0].start for x in rows)
            assert all(x.end == rows[0].end for x in rows)
            source = SeqRecord(id=rows[0].sequence_id, seq=_UnknownSeq())
            strand = Strand.Coding if rows[0].strand == "+" else Strand.Reverse
            protein = Protein(rows[0].protein_id, seq=None)
            gene = Gene(source, rows[0].start, rows[0].end, strand, protein)
            for row in rows:
                domain = Domain(row.domain, row.domain_start, row.domain_end, row.hmm, row.i_evalue, row.pvalue, row.bgc_probability)
                gene.protein.domains.append(domain)
            yield gene

    @requires("pandas")
    def to_dataframe(self) -> "pandas.DataFrame":
        """Convert the feature table to a `~pandas.DataFrame`.

        Raises:
            ImportError: if the `pandas` module could not be imported.

        """
        frame = pandas.DataFrame()  # type: ignore
        for column in self.__annotations__:
            frame[column] = getattr(self, column)
        return frame

    def __iadd__(self, rhs: "FeatureTable") -> "FeatureTable":  # noqa: D105
        if not isinstance(rhs, FeatureTable):
            return NotImplemented
        for col in self.__annotations__:
            getattr(self, col).extend(getattr(rhs, col))
        return self

    def __bool__(self) -> bool:  # noqa: D105
        return len(self) != 0

    def __len__(self) -> int:  # noqa: D105
        return len(self.sequence_id)

    def __iter__(self) -> Iterator[Row]:  # noqa: D105
        columns = { c: operator.attrgetter(c) for c in self.__annotations__ }
        for i in range(len(self)):
            row = { c: getter(self)[i] for c, getter in columns.items() }
            yield self.Row(**row)

    @typing.overload
    def __getitem__(self, item: slice) -> "FeatureTable":  # noqa: D105
        pass

    @typing.overload
    def __getitem__(self, item: int) -> Row:  # noqa: D105
        pass

    def __getitem__(self, item: Union[slice, int]) -> Union["FeatureTable", "Row"]:   # noqa: D105
        columns = [getattr(self, col)[item] for col in self.__annotations__]
        if isinstance(item, slice):
            return type(self)(*columns)
        else:
            return self.Row(*columns)

    def dump(self, fh: TextIO, dialect: str = "excel-tab", header: bool = True) -> None:
        """Write the feature table in CSV format to the given file.

        Arguments:
            fh (file-like `object`): A writable file-handle opened in text mode
                to write the feature table to.
            dialect (`str`): The CSV dialect to use. See `csv.list_dialects`
                for allowed values.
            header (`bool`): Whether or not to include the column header when
                writing the table (useful for appending to an existing table).
                Defaults to `True`.

        """
        writer = csv.writer(fh, dialect=dialect)
        columns = list(self.__annotations__)

        # do not write optional columns if they are completely empty
        if all(proba is None for proba in self.bgc_probability):
            columns.remove("bgc_probability")

        if header:
            writer.writerow(columns)
        for row in self:
            writer.writerow([ getattr(row, col) for col in columns ])

    @classmethod
    def load(cls, fh: TextIO, dialect: str = "excel-tab") -> "FeatureTable":
        """Load a feature table in CSV format from a file handle in text mode.
        """
        table = cls()
        reader = csv.reader(fh, dialect=dialect)
        header = next(reader)

        # get the name of each column
        columns = {i:col for i, col in enumerate(header)}

        # check that if a column is missing, it is one of the optional values
        missing = set(cls.__annotations__).difference(columns.values())
        missing_required = missing.difference({"bgc_probability"})
        if missing_required:
            raise ValueError("table is missing columns: {}".format(", ".join(missing_required)))

        # extract elements from the CSV rows
        for row in reader:
            for col in missing:
                getattr(table, col).append(None)
            for i,value in enumerate(row):
                col = columns.get(i)
                if col in ("i_evalue", "pvalue", "bgc_probability"):
                    getattr(table, col).append(float(value))
                elif col in ("start", "end", "domain_start", "domain_end"):
                    getattr(table, col).append(int(value))
                elif col in cls.__annotations__:
                    getattr(table, col).append(value)

        return table


@dataclass(frozen=True)
class ClusterTable(Dumpable, Sized):
    """A table storing condensed information from several clusters.
    """

    sequence_id: List[str] = field(default_factory = list)
    bgc_id: List[str] = field(default_factory = list)
    start: List[int] = field(default_factory = lambda: array("l"))          # type: ignore
    end: List[int] = field(default_factory = lambda: array("l"))            # type: ignore
    average_p: List[float] = field(default_factory = lambda: array("d"))    # type: ignore
    max_p: List[float] = field(default_factory = lambda: array("d"))        # type: ignore

    type: List[ProductType] = field(default_factory = list)
    alkaloid_probability: List[float] = field(default_factory = lambda: array("d"))    # type: ignore
    polyketide_probability: List[float] = field(default_factory = lambda: array("d"))  # type: ignore
    ripp_probability: List[float] = field(default_factory = lambda: array("d"))        # type: ignore
    saccharide_probability: List[float] = field(default_factory = lambda: array("d"))  # type: ignore
    terpene_probability: List[float] = field(default_factory = lambda: array("d"))     # type: ignore
    nrp_probability: List[float] = field(default_factory = lambda: array("d"))         # type: ignore
    other_probability: List[float] = field(default_factory = lambda: array("d"))       # type: ignore

    proteins: List[List[str]] = field(default_factory = list)
    domains: List[List[str]] = field(default_factory = list)

    class Row(NamedTuple):
        """A single row in a cluster table.
        """

        sequence_id: str
        bgc_id: str
        start: int
        end: int
        average_p: float
        max_p: float
        type: ProductType
        alkaloid_probability: float
        polyketide_probability: float
        ripp_probability: float
        saccharide_probability: float
        terpene_probability: float
        nrp_probability: float
        other_probability: float
        proteins: List[str]
        domains: List[str]

    @classmethod
    def from_clusters(cls, clusters: Iterable[Cluster]) -> "ClusterTable":
        """Create a new cluster table from an iterable of clusters.
        """
        table = cls()
        for cluster in clusters:
            table.sequence_id.append(cluster.source.id)
            table.bgc_id.append(cluster.id)
            table.start.append(cluster.start)
            table.end.append(cluster.end)
            table.average_p.append(cluster.average_probability)
            table.max_p.append(cluster.maximum_probability)

            table.type.append(cluster.type)
            table.alkaloid_probability.append(cluster.type_probabilities.get(ProductType.Alkaloid, 0))
            table.polyketide_probability.append(cluster.type_probabilities.get(ProductType.Polyketide, 0))
            table.ripp_probability.append(cluster.type_probabilities.get(ProductType.RiPP, 0))
            table.saccharide_probability.append(cluster.type_probabilities.get(ProductType.Saccharide, 0))
            table.terpene_probability.append(cluster.type_probabilities.get(ProductType.Terpene, 0))
            table.nrp_probability.append(cluster.type_probabilities.get(ProductType.NRP, 0))
            table.other_probability.append(cluster.type_probabilities.get(ProductType.Other, 0))

            table.proteins.append([ gene.protein.id for gene in cluster.genes ])
            domains = {d.name for g in cluster.genes for d in g.protein.domains}
            table.domains.append(sorted(domains))
        return table

    def __iadd__(self, rhs: "ClusterTable") -> "ClusterTable":  # noqa: D105
        if not isinstance(rhs, FeatureTable):
            return NotImplemented
        for col in self.__annotations__:
            getattr(self, col).extend(getattr(rhs, col))
        return self

    def __len__(self) -> int:  # noqa: D105
        return len(self.sequence_id)

    def __iter__(self) -> Iterator[Row]:  # noqa: D105
        columns = { c: operator.attrgetter(c) for c in self.__annotations__ }
        for i in range(len(self)):
            row = { c: getter(self)[i] for c, getter in columns.items() }
            yield self.Row(**row)

    def __bool__(self) -> bool:  # noqa: D105
        return len(self) != 0

    @typing.overload
    def __getitem__(self, item: slice) -> "ClusterTable":  # noqa: D105
        pass

    @typing.overload
    def __getitem__(self, item: int) -> Row:  # noqa: D105
        pass

    def __getitem__(self, item: Union[int, slice]) -> Union["ClusterTable", Row]:  # noqa: D105
        columns = [getattr(self, col)[item] for col in self.__annotations__]
        if isinstance(item, slice):
            return type(self)(*columns)
        else:
            return self.Row(*columns)

    def dump(self, fh: TextIO, dialect: str = "excel-tab", header: bool = True) -> None:
        """Write the cluster table in CSV format to the given file.

        Arguments:
            fh (file-like `object`): A writable file-handle opened in text mode
                to write the cluster table to.
            dialect (`str`): The CSV dialect to use. See `csv.list_dialects`
                for allowed values.
            header (`bool`): Whether or not to include the column header when
                writing the table (useful for appending to an existing table).
                Defaults to `True`.

        """
        writer = csv.writer(fh, dialect=dialect)
        columns = list(self.__annotations__)
        row = []
        if header:
            writer.writerow(columns)
        for i in range(len(self)):
            row.clear()
            for col in columns:
                value = getattr(self, col)[i]
                if col == "type":
                    types = value.unpack() or [ProductType.Unknown]
                    value = ";".join(map(operator.attrgetter("name"), types))
                elif isinstance(value, list):
                    value = ";".join(map(str, value))
                row.append(value)
            writer.writerow(row)

    @classmethod
    def load(cls, fh: TextIO, dialect: str = "excel-tab") -> "ClusterTable":
        """Load a cluster table in CSV format from a file handle in text mode.
        """
        table = cls()
        reader = csv.reader(fh, dialect=dialect)
        header = next(reader)

        # get the name of each column
        columns = {i:col for i, col in enumerate(header)}

        # check that if a column is missing, it is one of the optional values
        missing = set(cls.__annotations__).difference(columns.values())
        missing_required = missing.difference({
            "average_p",
            "max_p",
            "type",
            "alkaloid_probability",
            "polyketide_probability",
            "ripp_probability",
            "saccharide_probability",
            "terpene_probability",
            "nrp_probability",
            "other_probability",
            "proteins",
            "domains"
        })
        if missing_required:
            raise ValueError("table is missing columns: {}".format(", ".join(missing_required)))

        # extract elements from the CSV rows
        for row in reader:
            for col in missing:
                if col in ("proteins", "domains"):
                    getattr(table, col).append(list())
                elif col in ("average_p", "max_p"):
                    getattr(table, col).append(1.0)
                elif col == "type":
                    table.type.append(ProductType.Unknown)
                else:
                    getattr(table, col).append(0.0)
            for i,value in enumerate(row):
                col = columns.get(i)
                if col in ("i_evalue", "pvalue") or col.endswith(("_p", "_probability")):
                    getattr(table, col).append(float(value))
                elif col in ("start", "end", "domain_start", "domain_end"):
                    getattr(table, col).append(int(value))
                elif col == "type":
                    types = [ProductType.__members__[x] for x in value.split(";")]
                    table.type.append(ProductType.pack(types))
                elif col in cls.__annotations__:
                    getattr(table, col).append(value)

        return table



"""Implementation of the ``gecco run`` subcommand.
"""

import contextlib
import textwrap
from typing import Any, Dict, Mapping, List, Optional, TextIO

import rich.console
import rich.text

from .._utils import in_context, patch_showwarnings
from ._base import Command, CommandExit
from ._main import Main


class Help(Command):  # noqa: D101

    summary = "display the help message of another subcommand."

    @classmethod
    def doc(cls, fast=False):  # noqa: D102
        return f"""
        gecco help - {cls.summary}

        Usage:
            gecco help [<cmd>]

        Arguments:
            <cmd>                      a command to get the help message of.

        Parameters:
            -h, --help                 show the message for ``gecco`` or
                                       for a given subcommand.
        """

    def execute(self, ctx: contextlib.ExitStack) -> int:  # noqa: D102
        try:
            # check arguments and enter context
            self._check()
            ctx.enter_context(patch_showwarnings(self._showwarnings))
            # Get the subcommand class
            if self.args["<cmd>"] is not None:
                subcmd_cls = Main._get_subcommand_by_name(self.args["<cmd>"])
            else:
                subcmd_cls = None
            # Exit if no known command was found
            if self.args["<cmd>"] is not None and subcmd_cls is None:
                self.error("Unknown subcommand", repr(self.args["<cmd>"]))
                return 1
            # Render the help message
            doc = Main.doc() if subcmd_cls is None else subcmd_cls.doc()
            text = rich.text.Text(textwrap.dedent(doc).lstrip())
            console = rich.console.Console(file=self._stream, soft_wrap=True)
            console.print(text)
        except CommandExit as cexit:
            return cexit.code
        else:
            return 0



# coding: utf-8

import contextlib
import sys
import typing
from typing import Optional, List, TextIO

def main(argv: Optional[List[str]] = None, stream: Optional[TextIO] = None) -> int:
    from .commands._main import Main

    with contextlib.ExitStack() as ctx:
        return Main(argv, stream).execute(ctx)



"""Simple data classes to expose embedded InterPro data.
"""

import gzip
import json
from dataclasses import dataclass
from typing import Dict, List, Optional

try:
    import importlib.resources as importlib_resources
except ImportError:
    import importlib_resources  # type: ignore


__all__ = ["InterProEntry", "InterPro"]


@dataclass
class InterProEntry:
    """A single domain entry in the InterPro database.
    """

    accession: str
    go_terms: Dict[str, Dict[str, str]]
    integrated: Optional[str]
    member_databases: Dict[str, Dict[str, str]]
    name: str
    source_database: str
    type: str


@dataclass
class InterPro:
    """A subset of the InterPro database exposing domain metadata.
    """

    entries: List[InterProEntry]

    def __init__(self, entries: List[InterProEntry]):
        self.entries = entries
        self.by_accession = { entry.accession:entry for entry in entries }

    @classmethod
    def load(cls) -> "InterPro":
        with importlib_resources.open_binary(__name__, "interpro.json.gz") as f:
            data = json.load(gzip.open(f, mode="rt"))
            entries = [ InterProEntry(**entry["metadata"]) for entry in data ]
        return cls(entries)



"""Implementation of the ``gecco annotate`` subcommand.
"""

import contextlib
import errno
import glob
import gzip
import itertools
import io
import logging
import multiprocessing
import operator
import os
import pickle
import tempfile
import typing
import signal
from typing import Any, Dict, Union, Optional, List, TextIO, Mapping

from ._base import Command, CommandExit, InvalidArgument
from .._utils import (
    guess_sequences_format,
    in_context,
    patch_showwarnings,
    ProgressReader,
)


class Annotate(Command):  # noqa: D101

    summary = "annotate protein features of one or several contigs."

    @classmethod
    def doc(cls, fast=False):  # noqa: D102
        return f"""
        gecco annotate - {cls.summary}

        Usage:
            gecco annotate --genome <file> [--hmm <hmm>]... [options]

        Arguments:
            -g <file>, --genome <file>    a genomic file containing one or more
                                          sequences to use as input. Must be in
                                          one of the sequences format supported
                                          by Biopython.

        Parameters:
            -f <fmt>, --format <fmt>      the format of the input file, as a
                                          Biopython format string. GECCO is able
                                          to recognize FASTA and GenBank files
                                          automatically if this is not given.
            -o <out>, --output <out>      the file where to write the feature
                                          table. [default: features.tsv]
            -j <jobs>, --jobs <jobs>      the number of CPUs to use for
                                          multithreading. Use 0 to use all of the
                                          available CPUs. [default: 0]

        Parameters - Domain Annotation:
            -e <e>, --e-filter <e>        the e-value cutoff for protein domains
                                          to be included.
            -p <p>, --p-filter <p>        the p-value cutoff for protein domains
                                          to be included. [default: 1e-9]

        Parameters - Debug:
            --hmm <hmm>                   the path to one or more alternative
                                          HMM file to use (in HMMER format).
        """

    def _check(self) -> typing.Optional[int]:
        super()._check()
        try:
            self.e_filter = self._check_flag(
                "--e-filter",
                float,
                lambda x: x > 0,
                hint="real number above 0",
                optional=True,
            )
            self.p_filter = self._check_flag(
                "--p-filter",
                float,
                lambda x: x > 0,
                hint="real number above 0",
                optional=True,
            )
            self.jobs = self._check_flag("--jobs", int, lambda x: x >= 0, hint="positive or null integer")
            self.format = self._check_flag("--format")
            self.genome = self._check_flag("--genome")
            self.hmm = self._check_flag("--hmm")
            self.output = self._check_flag("--output")
        except InvalidArgument:
            raise CommandExit(1)

    def _custom_hmms(self):
        from ...hmmer import HMM

        for path in self.hmm:
            base = os.path.basename(path)
            file = open(path, "rb")
            if base.endswith(".gz"):
                base, _ = os.path.splitext(base)
                file = gzip.GzipFile(fileobj=file)
            base, _ = os.path.splitext(base)
            yield HMM(
                id=base,
                version="?",
                url="?",
                path=path,
                size=1,
                relabel_with=r"s/([^\.]*)(\..*)?/\1/"
            )

    # ---

    def _load_sequences(self):
        from Bio import SeqIO

        # guess format or use the one given in CLI
        if self.format is not None:
            format = self.format
            self.info("Using", "user-provided sequence format", repr(format), level=2)
        else:
            self.info("Detecting", "sequence format from file contents", level=2)
            format = guess_sequences_format(self.genome)
            self.success("Detected", "format of input as", repr(format), level=2)

        # get filesize and unit
        input_size = os.stat(self.genome).st_size
        total, scale, unit = ProgressReader.scale_size(input_size)
        task = self.progress.add_task("Loading sequences", total=total, unit=unit, precision=".1f")

        # load sequences
        self.info("Loading", "sequences from genomic file", repr(self.genome), level=1)
        try:
            with ProgressReader(open(self.genome, "rb"), self.progress, task, scale) as f:
                sequences = list(SeqIO.parse(io.TextIOWrapper(f), format))
        except FileNotFoundError as err:
            self.error("Could not find input file:", repr(self.genome))
            raise CommandExit(err.errno) from err
        except ValueError as err:
            self.error("Failed to load sequences:", err)
            raise CommandExit(getattr(err, "errno", 1)) from err
        else:
            self.success("Found", len(sequences), "sequences", level=1)
            return sequences

    def _extract_genes(self, sequences):
        from ...orf import PyrodigalFinder

        self.info("Extracting", "genes from input sequences", level=1)
        orf_finder = PyrodigalFinder(metagenome=True, cpus=self.jobs)

        unit = "contigs" if len(sequences) > 1 else "contig"
        task = self.progress.add_task(description="ORFs finding", total=len(sequences), unit=unit, precision="")

        def callback(record, found, total):
            self.success("Found", found, "genes in record", repr(record.id), level=2)
            self.progress.update(task, advance=1)

        return list(orf_finder.find_genes(sequences, progress=callback))

    def _annotate_domains(self, genes, whitelist=None):
        from ...hmmer import PyHMMER, embedded_hmms

        self.info("Running", "HMMER domain annotation", level=1)

        # Run all HMMs over ORFs to annotate with protein domains
        hmms = list(self._custom_hmms() if self.hmm else embedded_hmms())
        task = self.progress.add_task(description=f"HMM annotation", unit="HMMs", total=len(hmms), precision="")
        for hmm in self.progress.track(hmms, task_id=task, total=len(hmms)):
            task = self.progress.add_task(description=f"{hmm.id} v{hmm.version}", total=hmm.size, unit="domains", precision="")
            callback = lambda h, t: self.progress.update(task, advance=1)
            self.info("Starting", f"annotation with [bold blue]{hmm.id} v{hmm.version}[/]", level=2)
            features = PyHMMER(hmm, self.jobs, whitelist).run(genes, progress=callback)
            self.success("Finished", f"annotation with [bold blue]{hmm.id} v{hmm.version}[/]", level=2)
            self.progress.update(task_id=task, visible=False)

        # Count number of annotated domains
        count = sum(1 for gene in genes for domain in gene.protein.domains)
        self.success("Found", count, "domains across all proteins", level=1)

        # Filter i-evalue and p-value if required
        genes = self._filter_domains(genes)

        # Sort genes
        self.info("Sorting", "genes by coordinates", level=2)
        genes.sort(key=lambda g: (g.source.id, g.start, g.end))
        for gene in genes:
            gene.protein.domains.sort(key=operator.attrgetter("start", "end"))

        return genes

    def _filter_domains(self, genes):
        # Filter i-evalue and p-value if required
        if self.e_filter is not None:
            self.info("Filtering", "domains with e-value under", self.e_filter, level=1)
            key = lambda d: d.i_evalue < self.e_filter
            for gene in genes:
                gene.protein.domains = list(filter(key, gene.protein.domains))
        if self.p_filter is not None:
            self.info("Filtering", "domains with p-value under", self.p_filter, level=1)
            key = lambda d: d.pvalue < self.p_filter
            for gene in genes:
                gene.protein.domains = list(filter(key, gene.protein.domains))
        if self.p_filter is not None or self.e_filter is not None:
            count = sum(1 for gene in genes for domain in gene.protein.domains)
            self.info("Using", "remaining", count, "domains", level=1)
        return genes

    def _write_feature_table(self, genes):
        from ...model import FeatureTable

        self.info("Writing", "feature table to", repr(self.output), level=1)
        with open(self.output, "w") as f:
            FeatureTable.from_genes(genes).dump(f)

    # ---

    def execute(self, ctx: contextlib.ExitStack) -> int:  # noqa: D102
        try:
            # check the CLI arguments were fine and enter context
            self._check()
            ctx.enter_context(self.progress)
            ctx.enter_context(patch_showwarnings(self._showwarnings))
            # load sequences and extract genes
            sequences = self._load_sequences()
            genes = self._extract_genes(sequences)
            if genes:
                self.success("Found", "a total of", len(genes), "genes", level=1)
            else:
                self.warn("No genes were found")
                return 0
            # annotate domains and write results
            genes = self._annotate_domains(genes)
            self._write_feature_table(genes)
            ndoms = sum(1 for gene in genes for domain in gene.protein.domains)
            # report number of proteins found
            if ndoms:
                self.success("Found", ndoms, "protein domains", level=0)
            else:
                self.warn("No protein domains were found")
        except CommandExit as cexit:
            return cexit.code
        except KeyboardInterrupt:
            self.error("Interrupted")
            return -signal.SIGINT
        except Exception as err:
            self.progress.stop()
            raise
        else:
            return 0



"""Implementation of the ``gecco run`` subcommand.
"""

import contextlib
import errno
import glob
import itertools
import json
import logging
import multiprocessing
import operator
import os
import pickle
import tempfile
import typing
import signal
from typing import Any, Dict, Union, Optional, List, TextIO, Mapping

import rich.emoji
import rich.progress

from ... import __version__
from ._base import Command, CommandExit, InvalidArgument
from .annotate import Annotate
from .._utils import patch_showwarnings
from ...model import ProductType

try:
    import importlib.resources as importlib_resources
except ImportError:
    import importlib_resources


class Run(Annotate):  # noqa: D101

    summary = "predict BGC from one or several contigs."

    @classmethod
    def doc(cls, fast=False):  # noqa: D102
        return f"""
        gecco run - {cls.summary}

        Usage:
            gecco run --genome <file> [--hmm <hmm>]... [options]

        Arguments:
            -g <file>, --genome <file>    a genomic file containing one or more
                                          sequences to use as input. Must be in
                                          one of the sequences format supported
                                          by Biopython.

        Parameters:
            -f <fmt>, --format <fmt>      the format of the input file, as a
                                          Biopython format string. GECCO is able
                                          to recognize FASTA and GenBank files
                                          automatically if this is not given.
            -j <jobs>, --jobs <jobs>      the number of CPUs to use for
                                          multithreading. Use 0 to use all of the
                                          available CPUs. [default: 0]

        Parameters - Output:
            -o <out>, --output-dir <out>  the directory in which to write the
                                          output files. [default: .]
            --antismash-sideload          write an AntiSMASH v6 sideload JSON
                                          file next to the output files.

        Parameters - Domain Annotation:
            -e <e>, --e-filter <e>        the e-value cutoff for protein domains
                                          to be included.
            -p <p>, --p-filter <p>        the p-value cutoff for protein domains
                                          to be included. [default: 1e-9]

        Parameters - Cluster Detection:
            -c, --cds <N>                 the minimum number of coding sequences a
                                          valid cluster must contain. [default: 3]
            -m <m>, --threshold <m>       the probability threshold for cluster
                                          detection. Default depends on the
                                          post-processing method (0.4 for gecco,
                                          0.6 for antismash).
            --postproc <method>           the method to use for cluster validation
                                          (antismash or gecco). [default: gecco]

        Parameters - Debug:
            --model <directory>           the path to an alternative CRF model
                                          to use (obtained with `gecco train`).
            --hmm <hmm>                   the path to one or more alternative HMM
                                          file to use (in HMMER format).
        """

    def _check(self) -> typing.Optional[int]:
        Command._check(self)
        try:
            self.cds = self._check_flag("--cds", int, lambda x: x > 0, hint="positive integer")
            self.e_filter = self._check_flag(
                "--e-filter",
                float,
                lambda x: x > 0,
                hint="real number above 0",
                optional=True,
            )
            self.p_filter = self._check_flag(
                "--p-filter",
                float,
                lambda x: x > 0,
                hint="real number above 0",
                optional=True,
            )
            if self.args["--threshold"] is None:
                self.threshold = 0.4 if self.args["--postproc"] == "gecco" else 0.6
            else:
                self.threshold = self._check_flag("--threshold", float, lambda x: 0 <= x <= 1, hint="number between 0 and 1")
            self.jobs = self._check_flag("--jobs", int, lambda x: x >= 0, hint="positive or null integer")
            self.postproc = self._check_flag("--postproc", str, lambda x: x in ("gecco", "antismash"), hint="'gecco' or 'antismash'")
            self.format = self._check_flag("--format")
            self.genome = self._check_flag("--genome")
            self.model = self._check_flag("--model")
            self.hmm = self._check_flag("--hmm")
            self.output_dir = self._check_flag("--output-dir")
            self.antismash_sideload = self._check_flag("--antismash-sideload", bool)
        except InvalidArgument:
            raise CommandExit(1)

    # ---

    def _make_output_directory(self) -> None:
        # Make output directory
        self.info("Using", "output folder", repr(self.output_dir), level=1)
        try:
            os.makedirs(self.output_dir, exist_ok=True)
        except OSError as err:
            self.error("Could not create output directory: {}", err)
            raise CommandExit(err.errno) from err

        # Check if output files already exist
        base, _ = os.path.splitext(os.path.basename(self.genome))
        output_exts = ["features.tsv", "clusters.tsv"]
        if self.antismash_sideload:
            output_exts.append("sideload.json")
        for ext in output_exts:
            if os.path.isfile(os.path.join(self.output_dir, f"{base}.{ext}")):
                self.warn("Output folder contains files that will be overwritten")
                break

    def _load_model_domains(self) -> typing.Set[str]:
        try:
            if self.model is None:
                self.info("Loading", "feature list from internal model", level=2)
                domains_file = importlib_resources.open_text("gecco.types", "domains.tsv")
            else:
                self.info("Loading", "feature list from", repr(self.model), level=2)
                domains_file = open(os.path.join(self.model, "domains.tsv"))
            with domains_file as f:
                domains = set(filter(None, map(str.strip, f)))
        except FileNotFoundError as err:
            if self.model is not None:
                self.error("Could not find domains list :", repr(self.model))
            raise CommandExit(e.errno) from err
        else:
            self.success("Found", len(domains), "selected features", level=2)
            return domains

    def _predict_probabilities(self, genes):
        from ...crf import ClusterCRF

        if self.model is None:
            self.info("Loading", "embedded CRF pre-trained model", level=1)
        else:
            self.info("Loading", "CRF pre-trained model from", repr(self.model), level=1)
        crf = ClusterCRF.trained(self.model)

        self.info("Predicting", "cluster probabilitites with the CRF model", level=1)
        unit = "genes" if len(genes) > 1 else "gene"
        task = self.progress.add_task("Prediction", total=len(genes), unit=unit, precision="")
        return list(crf.predict_probabilities(
            self.progress.track(genes, task_id=task, total=len(genes)),
            cpus=self.jobs
        ))

    def _write_feature_table(self, genes):
        from ...model import FeatureTable

        base, _ = os.path.splitext(os.path.basename(self.genome))
        pred_out = os.path.join(self.output_dir, f"{base}.features.tsv")
        self.info("Writing", "feature table to", repr(pred_out), level=1)
        with open(pred_out, "w") as f:
            FeatureTable.from_genes(genes).dump(f)

    def _extract_clusters(self, genes):
        from ...refine import ClusterRefiner

        self.info("Extracting", "predicted biosynthetic regions", level=1)
        refiner = ClusterRefiner(self.threshold, self.postproc, self.cds)

        total = len({gene.source.id for gene in genes})
        unit = "contigs" if total > 1 else "contig"
        task = self.progress.add_task("Segmentation", total=total, unit=unit, precision="")

        clusters = []
        gene_groups = itertools.groupby(genes, lambda g: g.source.id)
        for _, gene_group in self.progress.track(gene_groups, task_id=task, total=total):
            clusters.extend(refiner.iter_clusters(gene_group))

        return clusters

    def _predict_types(self, clusters):
        from ...model import ProductType
        from ...types import TypeClassifier

        self.info("Predicting", "BGC types", level=1)

        unit = "cluster" if len(clusters) == 1 else "clusters"
        task = self.progress.add_task("Type prediction", total=len(clusters), unit=unit, precision="")

        clusters_new = []
        classifier = TypeClassifier.trained(self.model)
        for cluster in self.progress.track(clusters, task_id=task):
            clusters_new.extend(classifier.predict_types([cluster]))
            if cluster.type != ProductType.Unknown:
                name = "/".join(f"[bold blue]{ty.name}[/]" for ty in cluster.type.unpack())
                prob = "/".join(f"[bold purple]{cluster.type_probabilities[ty]:.0%}[/]" for ty in cluster.type.unpack())
                self.success(f"Predicted type of [bold blue]{cluster.id}[/] as {name} ({prob} confidence)")
            else:
                ty = max(cluster.type_probabilities, key=cluster.type_probabilities.get)
                prob = f"[bold purple]{cluster.type_probabilities[ty]:.0%}[/]"
                name = f"[bold blue]{ty.name}[/]"
                self.warn(f"Couldn't assign type to [bold blue]{cluster.id}[/] (maybe {name}, {prob} confidence)")

        return clusters_new

    def _write_cluster_table(self, clusters):
        from ...model import ClusterTable

        base, _ = os.path.splitext(os.path.basename(self.genome))
        cluster_out = os.path.join(self.output_dir, f"{base}.clusters.tsv")
        self.info("Writing", "cluster table to", repr(cluster_out), level=1)
        with open(cluster_out, "w") as out:
            ClusterTable.from_clusters(clusters).dump(out)

    def _write_clusters(self, clusters):
        from Bio import SeqIO

        for cluster in clusters:
            gbk_out = os.path.join(self.output_dir, f"{cluster.id}.gbk")
            self.info("Writing", f"cluster [bold blue]{cluster.id}[/] to", repr(gbk_out), level=1)
            SeqIO.write(cluster.to_seq_record(), gbk_out, "genbank")

    def _write_sideload_json(self, clusters):
        # record version and important parameters
        data = {
            "records": [],
            "tool": {
                "name": "GECCO",
                "version": __version__,
                "description": "Biosynthetic Gene Cluster prediction with Conditional Random Fields.",
                "configuration": {
                    "cds": repr(self.cds),
                    "e-filter": repr(self.e_filter),
                    "postproc": repr(self.postproc),
                    "threshold": repr(self.threshold),
                }
            }
        }
        # record if non-standard HMM or model was used
        if self.hmm:
            data["tool"]["configuration"]["hmm"] = list(self.hmm)
        if self.model:
            data["tool"]["configuration"]["model"] = self.model
        # create a record per sequence
        for seq_id, seq_clusters in itertools.groupby(clusters, key=lambda cluster: cluster.source.id):
            data["records"].append({"name": seq_id, "subregions": []})
            for cluster in seq_clusters:
                ty = ";".join(sorted(ty.name for ty in cluster.type.unpack())) or "Unknown"
                data["records"][-1]["subregions"].append({
                    "start": cluster.start,
                    "end": cluster.end,
                    "label": ty,
                    "details": {
                        "average_p": f"{cluster.average_probability:.3f}",
                        "max_p": f"{cluster.maximum_probability:.3f}",
                        "alkaloid_probability": f"{cluster.type_probabilities.get(ProductType.Alkaloid, 0.0):.3f}",
                        "polyketide_probability": f"{cluster.type_probabilities.get(ProductType.Polyketide, 0.0):.3f}",
                        "ripp_probability": f"{cluster.type_probabilities.get(ProductType.RiPP, 0.0):.3f}",
                        "saccharide_probability": f"{cluster.type_probabilities.get(ProductType.Saccharide, 0.0):.3f}",
                        "terpene_probability": f"{cluster.type_probabilities.get(ProductType.Terpene, 0.0):.3f}",
                        "nrp_probability": f"{cluster.type_probabilities.get(ProductType.NRP, 0.0):.3f}",
                        "other_probability": f"{cluster.type_probabilities.get(ProductType.Other, 0.0):.3f}",
                    }
                })
        # write the JSON file to the output folder
        base, _ = os.path.splitext(os.path.basename(self.genome))
        sideload_out = os.path.join(self.output_dir, f"{base}.sideload.json")
        self.info("Writing", "sideload JSON to", repr(sideload_out), level=1)
        with open(sideload_out, "w") as out:
            json.dump(data, out, sort_keys=True, indent=4)

    # ---

    def execute(self, ctx: contextlib.ExitStack) -> int:  # noqa: D102
        try:
            # check the CLI arguments were fine and enter context
            self._check()
            ctx.enter_context(self.progress)
            ctx.enter_context(patch_showwarnings(self._showwarnings))
            # attempt to create the output directory
            self._make_output_directory()
            # load sequences and extract genes
            sequences = self._load_sequences()
            genes = self._extract_genes(sequences)
            if genes:
                self.success("Found", "a total of", len(genes), "genes", level=1)
            else:
                self.warn("No genes were found")
                return 0
            # use a whitelist for domain annotation, so that we only annotate
            # with features that are useful for the CRF
            whitelist = self._load_model_domains()
            # annotate domains and predict probabilities
            genes = self._annotate_domains(genes, whitelist=whitelist)
            genes = self._predict_probabilities(genes)
            self._write_feature_table(genes)
            # extract clusters from probability vector
            clusters = self._extract_clusters(genes)
            if clusters:
                self.success("Found", len(clusters), "potential gene clusters", level=1)
            else:
                self.warn("No gene clusters were found")
                return 0
            # predict types for putative clusters
            clusters = self._predict_types(clusters)
            # write results
            self.info("Writing", "result files to folder", repr(self.output_dir), level=1)
            self._write_cluster_table(clusters)
            self._write_clusters(clusters)
            if self.antismash_sideload:
                self._write_sideload_json(clusters)
            unit = "cluster" if len(clusters) == 1 else "clusters"
            self.success("Found", len(clusters), "biosynthetic gene", unit, level=0)
        except CommandExit as cexit:
            return cexit.code
        except KeyboardInterrupt:
            self.error("Interrupted")
            return -signal.SIGINT
        except Exception as err:
            self.progress.stop()
            raise
        else:
            return 0



"""Implementation of the ``gecco convert`` subcommand.
"""

import contextlib
import copy
import errno
import functools
import itertools
import glob
import os
import operator
import multiprocessing
import random
import typing
from typing import Any, Dict, Union, Optional, List, TextIO, Mapping

from ... import __version__
from ...model import ClusterTable
from ._base import Command, CommandExit, InvalidArgument
from .._utils import patch_showwarnings


class Convert(Command):  # noqa: D101

    summary = "convert output for compatibility with other tools"

    @classmethod
    def doc(cls, fast=False):  # noqa: D102
        return f"""
        gecco convert  - {cls.summary}

        Usage:
            gecco convert gbk -i <input> -f <format> [options]

        Arguments:
            -i <input>, --input-dir <input> the path to the input directory
                                            containing files to convert.
            -f <format>, --format <format>  the name of the output format to
                                            write.

        Parameters:
            -o <out>, --output <out>        the name of the output file,
                                            default depends on the output
                                            format.

        This command helps transforming the output files created by
        GECCO into helpful format, should you want to use the results in
        combination with other tools. The supported formats are listed
        below:

        * ``gecco convert gbk --format=bigslice``: convert and alias the
          GenBank files in the given directory so that they can be loaded by
          BiG-SLiCE. Output files are named ``*.regionNNN.gbk``.
        * ``gecco convert gbk --format=fna``: convert the GenBank files to
          FASTA files containing the nucleotide sequence of the cluster.
          Output files are named ``*.fna``.
        * ``gecco convert gbk --format=faa``: convert the GenBank files to
          FASTA files containing the amino-acid sequences of all the proteins
          in a cluster. Output files are named ``*.faa``.

        """

    _CLUSTERS_FORMATS = set()
    _GBK_FORMATS = {"bigslice", "faa", "fna"}

    def _check(self) -> typing.Optional[int]:
        try:
            super()._check()
            formats = self._GBK_FORMATS if self.args["gbk"] else self._CLUSTERS_FORMATS
            self.format = self._check_flag("--format", str, lambda x: x in formats, hint=", ".join(formats))
            self.input_dir = self._check_flag("--input-dir")
        except InvalidArgument:
            raise CommandExit(1)

    # ---

    def _convert_gbk_bigslice(self, ctx: contextlib.ExitStack):
        import Bio.SeqIO
        from Bio.SeqFeature import SeqFeature, FeatureLocation

        # locate original folder
        if not os.path.exists(self.input_dir):
            self.error("Could not find input folder:", repr(self.input_dir))
            raise CommandExit(errno.ENOENT)
        # collect `*.clusters.tsv` files
        cluster_files = glob.glob(os.path.join(self.input_dir, "*.clusters.tsv"))
        unit = "table" if len(cluster_files) == 1 else "tables"
        task = self.progress.add_task("Loading", total=len(cluster_files), unit=unit, precision="")
        # load the original coordinates from the `*.clusters.tsv` files
        coordinates = {}
        types = {}
        for cluster_file in self.progress.track(cluster_files, task_id=task, precision=""):
            cluster_fh = ctx.enter_context(open(cluster_file))
            for row in ClusterTable.load(cluster_fh):
                ty = ";".join(sorted(ty.name for ty in row.type.unpack()))
                coordinates[row.bgc_id] = (row.start, row.end)
                types[row.bgc_id] = ty or "Unknown"

        # collect `*_clusters_{N}.gbk` files
        gbk_files = glob.glob(os.path.join(self.input_dir, "*_cluster_*.gbk"))
        unit = "file" if len(gbk_files) == 1 else "files"
        task = self.progress.add_task("Converting", total=len(gbk_files), unit=unit, precision="")
        done = 0
        # rewrite GenBank files
        for gbk_file in self.progress.track(gbk_files, task_id=task, total=len(gbk_files), precision=""):
            # load record and ensure it comes from GECCO
            record = Bio.SeqIO.read(gbk_file, "genbank")
            if "GECCO-Data" not in record.annotations.get('structured_comment', {}):
                self.warning(f"GenBank file {gbk_file!r} was not obtained by GECCO")
                continue
            # mark as AntiSMASH v5
            record.annotations['structured_comment']['antiSMASH-Data'] = {
                "Version": "5.X",
                "Orig. start": coordinates[record.id][0],
                "Orig. end": coordinates[record.id][1],
            }
            # using a /subregion feature will make BiG-SLiCE think the
            # BGC is coming from MIBiG, and the predicted type will be
            # handled correctly
            subregion_feature = SeqFeature(FeatureLocation(0, len(record)), type="subregion")
            subregion_feature.qualifiers["contig_edge"] = ["False"]
            subregion_feature.qualifiers["aStool"] = ["mibig"]
            subregion_feature.qualifiers["label"] = [types[record.id]]
            record.features.append(subregion_feature)
            # rename the {id}_cluster_{N}.gbk file to {id}.region{N}.gbk
            new_name = gbk_file.replace("_cluster_", ".region")
            self.info(f"Rewriting {gbk_file!r} to {new_name!r}")
            Bio.SeqIO.write(record, new_name, "genbank")
            done += 1

        self.success(f"Converted {done} GenBank {unit} to BiG-SLiCE format", level=0)

    def _convert_gbk_fna(self, ctx: contextlib.ExitStack):
        import Bio.SeqIO
        from Bio.SeqFeature import SeqFeature, FeatureLocation

        # collect `*_clusters_{N}.gbk` files
        gbk_files = glob.glob(os.path.join(self.input_dir, "*_cluster_*.gbk"))
        unit = "file" if len(gbk_files) == 1 else "files"
        task = self.progress.add_task("Converting", total=len(gbk_files), unit=unit, precision="")
        done = 0
        # rewrite GenBank files
        for gbk_file in self.progress.track(gbk_files, task_id=task, total=len(gbk_files)):
            # load record and ensure it comes from GECCO
            record = Bio.SeqIO.read(gbk_file, "genbank")
            if "GECCO-Data" not in record.annotations.get('structured_comment', {}):
                self.warning(f"GenBank file {gbk_file!r} was not obtained by GECCO")
                continue
            # convert to nucleotide FASTA
            new_name = gbk_file.replace(".gbk", ".fna")
            self.info(f"Converting {gbk_file!r} to FASTA file {new_name!r}")
            Bio.SeqIO.write(record, new_name, "fasta")
            done += 1
        self.success(f"Converted {done} GenBank {unit} to nucleotide FASTA format", level=0)

    def _convert_gbk_faa(self, ctx: contextlib.ExitStack):
        import Bio.SeqIO
        from Bio.Seq import Seq
        from Bio.SeqRecord import SeqRecord
        from Bio.SeqFeature import SeqFeature, FeatureLocation

        # collect `*_clusters_{N}.gbk` files
        gbk_files = glob.glob(os.path.join(self.input_dir, "*_cluster_*.gbk"))
        unit = "file" if len(gbk_files) == 1 else "files"
        task = self.progress.add_task("Converting", total=len(gbk_files), unit=unit, precision="")
        done = 0
        # rewrite GenBank files
        for gbk_file in self.progress.track(gbk_files, task_id=task, total=len(gbk_files)):
            # load record and ensure it comes from GECCO
            record = Bio.SeqIO.read(gbk_file, "genbank")
            if "GECCO-Data" not in record.annotations.get('structured_comment', {}):
                self.warning(f"GenBank file {gbk_file!r} was not obtained by GECCO")
                continue
            # extract proteins
            proteins = []
            for feat in filter(lambda f: f.type == "CDS", record.features):
                if "locus_tag" not in feat.qualifiers:
                    continue
                seq = Seq(feat.qualifiers["translation"][0])
                prot = SeqRecord(id=feat.qualifiers["locus_tag"][0], seq=seq)
                proteins.append(prot)
            # write proteins to FASTA
            new_name = gbk_file.replace(".gbk", ".faa")
            self.info(f"Converting {gbk_file!r} proteins to FASTA file {new_name!r}")
            Bio.SeqIO.write(proteins, new_name, "fasta")
            done += 1
        self.success(f"Converted {done} GenBank {unit} to protein FASTA format", level=0)

    def execute(self, ctx: contextlib.ExitStack) -> int:  # noqa: D102
        try:
            # check the CLI arguments were fine and enter context
            self._check()
            ctx.enter_context(self.progress)
            ctx.enter_context(patch_showwarnings(self._showwarnings))
            # run the appropriate method
            if self.args["gbk"]:
                if self.args["--format"] == "bigslice":
                    self._convert_gbk_bigslice(ctx)
                elif self.args["--format"] == "fna":
                    self._convert_gbk_fna(ctx)
                elif self.args["--format"] == "faa":
                    self._convert_gbk_faa(ctx)
        except CommandExit as cexit:
            return cexit.code
        except KeyboardInterrupt:
            self.error("Interrupted")
            return -signal.SIGINT
        except Exception as err:
            self.progress.stop()
            raise
        else:
            return 0
